#pragma once

#include "nix/store/parsed-derivations.hh"
#include "nix/store/store-api.hh"
#include "nix/store/pathlocks.hh"
#include "nix/store/build/goal.hh"

namespace nix {

/**
 * This goal type is essentially the serial composition (like function
 * composition) of a goal for getting a derivation, and then a
 * `DerivationGoal` using the newly-obtained derivation.
 *
 * In the (currently experimental) general inductive case of derivations
 * that are themselves build outputs, that first goal will be *another*
 * `DerivationCreationAndRealisationGoal`. In the (much more common) base-case
 * where the derivation has no provence and is just referred to by
 * (content-addressed) store path, that first goal is a
 * `SubstitutionGoal`.
 *
 * If we already have the derivation (e.g. if the evaluator has created
 * the derivation locally and then instructured the store to build it),
 * we can skip the first goal entirely as a small optimization.
 */
struct DerivationCreationAndRealisationGoal : public Goal
{
    /**
     * How to obtain a store path of the derivation to build.
     */
    ref<const SingleDerivedPath> drvReq;

    /**
     * The path of the derivation, once obtained.
     **/
    std::optional<StorePath> optDrvPath;

    /**
     * The specific outputs that we need to build.
     */
    OutputsSpec wantedOutputs;

    /**
     * The final output paths of the build.
     *
     * - For input-addressed derivations, always the precomputed paths
     *
     * - For content-addressing derivations, calcuated from whatever the
     *   hash ends up being. (Note that fixed outputs derivations that
     *   produce the "wrong" output still install that data under its
     *   true content-address.)
     */
    OutputPathMap finalOutputs;

    BuildMode buildMode;

    DerivationCreationAndRealisationGoal(
        ref<const SingleDerivedPath> drvReq,
        const OutputsSpec & wantedOutputs,
        Worker & worker,
        BuildMode buildMode = bmNormal);

    DerivationCreationAndRealisationGoal(
        const StorePath & drvPath,
        const OutputsSpec & wantedOutputs,
        const Derivation & drv,
        Worker & worker,
        BuildMode buildMode = bmNormal);

    virtual ~DerivationCreationAndRealisationGoal();

    void timedOut(Error && ex) override;

    std::string key() override;

    Co init();
    Co haveDerivation(StorePath drvPath, Derivation drv);

    JobCategory jobCategory() const override
    {
        return JobCategory::Administration;
    };

private:
    /**
     * Shared between both constructors
     */
    void commonInit();
};

}
